'''
This module contains functions to generate 2D rotation matrices.
'''
import math
import numpy as np

def rotation_from_angle(angle):
    '''
    generates a 2x2 rotation matrix from a given angle following
    the definition here:
    https://en.wikipedia.org/wiki/Rotation_matrix.

    :param angle: the angle by which to rotate
    :type angle: float

    :returns: the rotation matrix
    :rtype: a 2x2 :any:`numpy.ndarray`
    '''
    c = math.cos(angle)
    s = math.sin(angle)

    return np.array([[c, -s],[s, c]])



from .rotation_matrix_3d import rotation_from_angles
from .rotation_matrix_3d import rotation_around_x
from .rotation_matrix_3d import rotation_around_y
from .rotation_matrix_3d import rotation_around_z
from .rotation_matrix_3d import rotation_around_axis
from .rotation_matrix_nd import rotation_from_angle_and_plane
from .rotation_matrix_2d import rotation_from_angle
from .rotation_matrix_nd import random_matrix

from pbr.version import VersionInfo

_v = VersionInfo('mgen').semantic_version()
__version__ = _v.release_string()
version_info = _v.version_tuple()



'''
This module contains functions to generate 3D rotation matrices.
'''

import numpy as np
import math

def rotation_around_axis(axis, angle):
    '''
    Generates a 3x3 rotation matrix using the Euler-Rodrigues formula
    following the definition here:
    https://en.wikipedia.org/wiki/Euler%E2%80%93Rodrigues_formula.

    :param axis: the axis around which to rotate as a vector of length 3
                 (no normalisation required)
    :type axis: array like
    :param angle: the angle in radians to rotate
    :type angle: float

    :returns: the rotation matrix
    :rtype: a 3x3 :any:`numpy.ndarray`
    '''
    axis = axis/np.linalg.norm(axis)

    a = math.cos(angle/2.0)
    b, c, d = axis * math.sin(angle/2.0)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d

    return np.array([[aa+bb-cc-dd, 2*(bc-ad), 2*(bd+ac)],
                     [2*(bc+ad), aa+cc-bb-dd, 2*(cd-ab)],
                     [2*(bd-ac), 2*(cd+ab), aa+dd-bb-cc]])


def rotation_around_x(angle):
    '''
    Generates a 3x3 rotation matrix that performs a rotation around
    the x-axis following the definitions here:
    https://en.wikipedia.org/wiki/Rotation_matrix#Basic_rotations

    :param angle: the angle by which to rotate around the x-axis
    :type angle: float

    :returns: the rotation matrix
    :rtype: a 3x3 :any:`numpy.ndarray`
    '''
    c = math.cos(angle)
    s = math.sin(angle)

    return np.array(
        [[1, 0, 0],
         [0, c, -s],
         [0, s, c]])

def rotation_around_y(angle):
    '''
    Generates a 3x3 rotation matrix that performs a rotation around
    the y-axis following the definitions here:
    https://en.wikipedia.org/wiki/Rotation_matrix#Basic_rotations

    :param angle: the angle by which to rotate around the y-axis
    :type angle: float

    :returns: the rotation matrix
    :rtype: a 3x3 :any:`numpy.ndarray`
    '''
    c = math.cos(angle)
    s = math.sin(angle)

    return np.array(
        [[c, 0, s],
         [0, 1, 0],
         [-s, 0, c]])

def rotation_around_z(angle):
    '''
    Generates a 3x3 rotation matrix that performs a rotation around
    the z-axis following the definitions here:
    https://en.wikipedia.org/wiki/Rotation_matrix#Basic_rotations

    :param angle: the angle by which to rotate around the z-axis
    :type angle: float

    :returns: the rotation matrix
    :rtype: a 3x3 :any:`numpy.ndarray`
    '''
    c = math.cos(angle)
    s = math.sin(angle)

    return np.array(
        [[c, -s, 0],
         [s, c, 0],
         [0, 0, 1]])


# Implementation of the different variations of rotation matrices as
# described in https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix.

def _generate_matrix_XZX(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[ c2                , -c3*s2            ,  s2*s3            ],
         [ c1*s2             ,  c1*c2*c3 - s1*s3 , -c3*s1 - c1*c2*s3 ],
         [ s1*s2             ,  c1*s3 + c2*c3*s1 ,  c1*c3 - c2*s1*s3 ]])

def _generate_matrix_XYX(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[ c2                ,  s2*s3            ,  c3*s2            ],
         [ s1*s2             ,  c1*c3 - c2*s1*s3 , -c1*s3 - c2*c3*s1 ],
         [ -c1*s2            ,  c3*s1 + c1*c2*s3 ,  c1*c2*c3 - s1*s3 ]])

def _generate_matrix_YXY(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c3 - c2*s1*s3 ,  s1*s2            ,  c1*s3 + c2*c3*s1 ],
         [  s2*s3            ,  c2               , -c3*s2            ],
         [  -c3*s1 -c1*c2*s3 ,  c1*s2            ,  c1*c2*c3 - s1*s3 ]])

def _generate_matrix_YZY(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c2*c3 - s1*s3 , -c1*s2            ,  c3*s1 + c1*c2*s3 ],
         [  c3*s2            ,  c2               ,  s2*s3            ],
         [ -c1*s3 - c2*c3*s1 ,  s1*s2            ,  c1*c3 - c2*s1*s3 ]])

def _generate_matrix_ZYZ(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c2*c3 - s1*s3 , -c3*s1 - c1*c2*s3 ,  c1*s2            ],
         [  c1*s3 + c2*c3*s1 ,  c1*c3 - c2*s1*s3 ,  s1*s2            ],
         [  -c3*s2           ,  s2*s3            ,  c2               ]])

def _generate_matrix_ZXZ(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c3 - c2*s1*s3 , -c1*s3 - c2*c3*s1 ,  s1*s2            ],
         [  c3*s1 + c1*c2*s3 ,  c1*c2*c3 - s1*s3 , -c1*s2            ],
         [  s2*s3            ,  c3*s2            ,  c2               ]])

def _generate_matrix_XZY(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c2*c3            , -s2               ,  c2*s3            ],
         [  s1*s3 + c1*c3*s2 ,  c1*c2            ,  c1*s2*s3 - c3*s1 ],
         [  c3*s1*s2 - c1*s3 ,  c2*s1            ,  c1*c3 + s1*s2*s3 ]])

def _generate_matrix_XYZ(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c2*c3            , -c2*s3            ,  s2               ],
         [  c1*s3 + c3*s1*s2 ,  c1*c3 - s1*s2*s3 , -c2*s1            ],
         [  s1*s3 - c1*c3*s2 ,  c3*s1 + c1*s2*s3 ,  c1*c2            ]])

def _generate_matrix_YXZ(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c3 + s1*s2*s3 ,  c3*s1*s2 - c1*s3 ,  c2*s1            ],
         [  c2*s3            ,  c2*c3            , -s2               ],
         [  c1*s2*s3 - c3*s1 ,  c1*c3*s2 + s1*s3 ,  c1*c2            ]])

def _generate_matrix_YZX(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c2            ,  s1*s3 - c1*c3*s2 ,  c3*s1 + c1*s2*s3 ],
         [  s2               ,  c2*c3            ,  -c2*s3           ],
         [ -c2*s1            ,  c1*s3 + c3*s1*s2 ,  c1*c3 - s1*s2*s3 ]])

def _generate_matrix_ZYX(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c2            ,  c1*s2*s3 - c3*s1 ,  s1*s3 + c1*c3*s2 ],
         [  c2*s1            ,  c1*c3 + s1*s2*s3 ,  c3*s1*s2 - c1*s3 ],
         [ -s2               ,  c2*s3            ,  c2*c3            ]])

def _generate_matrix_ZXY(c1, c2, c3, s1, s2, s3):
    return np.asarray(
        [[  c1*c3 - s1*s2*s3 , -c2*s1            ,  c1*s3 + c3*s1*s2 ],
         [  c3*s1 + c1*s2*s3 ,  c1*c2            ,  s1*s3 - c1*c3*s2 ],
         [ -c2*s3            ,  s2               ,  c2*c3            ]])

# End of wikipedia copy paste


def rotation_from_angles(angles, rotation_sequence):
    '''
    Generate a 3x3 rotation matrix using proper Euler angles or
    Tait-Bryan angles as defined here: https://en.wikipedia.org/wiki/Euler_angles#Rotation_matrix.
    The angles given correspond to rotations as given in rotation_sequence, e.g.::

        rotation_from_angles([np.pi/2, 2/3*np.pi, np.pi/3], 'ZYX')

    would create a rotation matrix equal to the product of a rotation around Z by 90° times
    a rotation around Y by 60° times a rotation around X by 30°. For all the details
    please have a look at the linked wikipedia article.


    :param angles: the three angles in radians that define the rotation as a vector
                   of length 3
    :type angles: array like
    :param rotation_sequence: the sequence of rotations that make up the
                              total rotation. Example: `XYZ` yields the rotation
                              matrix :math:`R=XYZ`, i.e. the product of the
                              three matrices :math:`X`, :math:`Y` and :math:`Z`.
    :type rotation_sequence: str
    :returns: the rotation matrix
    :rtype: a 3x3 :any:`numpy.array`
    '''

    c1, c2, c3 = np.cos(angles)
    s1, s2, s3 = np.sin(angles)
    try:
        return globals()['_generate_matrix_'+rotation_sequence](c1, c2, c3, s1, s2, s3)
    except KeyError:
        raise ValueError(
            'Sequence ' + rotation_sequence
            + ' is not valid.')



'''
This module contains functions to generate n-D rotation matrices.
'''
import math
import numpy as np

def rotation_from_angle_and_plane(angle, vector1, vector2, abs_tolerance=1e-10):
    '''
    generates an nxn rotation matrix from a given angle and
    a plane spanned by two given vectors of length n:
    https://de.wikipedia.org/wiki/Drehmatrix#Drehmatrizen_des_Raumes_%7F'%22%60UNIQ--postMath-0000003B-QINU%60%22'%7F

    The formula used is

    .. math::

        M = 𝟙 + (\cos\\alpha-1)\cdot(v_1\otimes v_1 + v_2\otimes v_2) - \sin\\alpha\cdot(v_1\otimes v_2 - v_2\otimes v_1)

    with :math:`M` being the returned matrix, :math:`v_1` and :math:`v_2` being the two
    given vectors and :math:`\\alpha` being the given angle. It differs from the formula
    on wikipedia in that it is the transposed matrix to yield results that are consistent
    with the 2D and 3D cases.

    :param angle: the angle by which to rotate
    :type angle: float
    :param vector1: one of the two vectors that span the plane in which to rotate
                    (no normalisation required)
    :type vector1: array like
    :param vector2: the other of the two vectors that span the plane in which to rotate
                    (no normalisation required)
    :type vector2: array like
    :param abs_tolerance: The absolute tolerance to use when checking if vectors have length 0 or are parallel.
    :type abs_tolerance: float

    :returns: the rotation matrix
    :rtype: an nxn :any:`numpy.ndarray`
    '''

    vector1 = np.asarray(vector1, dtype=float)
    vector2 = np.asarray(vector2, dtype=float)

    vector1_length = np.linalg.norm(vector1)
    if math.isclose(vector1_length, 0., abs_tol=abs_tolerance):
        raise ValueError(
            'Given vector1 must have norm greater than zero within given numerical tolerance: {:.0e}'.format(abs_tolerance))

    vector2_length = np.linalg.norm(vector2)
    if math.isclose(vector2_length, 0., abs_tol=abs_tolerance):
        raise ValueError(
            'Given vector2 must have norm greater than zero within given numerical tolerance: {:.0e}'.format(abs_tolerance))

    vector2 /= vector2_length
    dot_value = np.dot(vector1, vector2)

    if abs(dot_value / vector1_length ) > 1 - abs_tolerance:
        raise ValueError(
            'Given vectors are parallel within the given tolerance: {:.0e}'.format(abs_tolerance))

    if abs(dot_value / vector1_length ) > abs_tolerance:
        vector1 = vector1 - dot_value * vector2
        vector1 /= np.linalg.norm(vector1)
    else:
        vector1 /= vector1_length


    vectors = np.vstack([vector1, vector2]).T
    vector1, vector2 = np.linalg.qr(vectors)[0].T

    V = np.outer(vector1, vector1) + np.outer(vector2, vector2)
    W = np.outer(vector1, vector2) - np.outer(vector2, vector1)

    return np.eye(len(vector1)) + (math.cos(angle) - 1)*V - math.sin(angle)*W





def random_matrix(n):
    '''
    Generate a nxn random matrix.
    The distribution of rotations is uniform on the n-sphere.
    The random matrix is from the O(n) group (not SO(n)) and uses the
    gram-schmidt algorithm to orthogonalize the random matrix,
    see http://www.ams.org/notices/200511/what-is.pdf
    If a rotation from SO(n) is needed, look for:
    "A statistical model for random rotations" doi:10.1016/j.jmva.2005.03.009

    :param n: dimension of space in which the rotation operates
    :type n: int
    :returns: rotation matrix
    :rtype: a nxn :any:`numpy.ndarray`
    '''
    if n < 2:
        raise ValueError('n must be greater than 1 but is ' + str(n) + '.')

    a = np.random.randn(n,n)

    a[0] = a[0]/np.linalg.norm(a[0])

    max_iter = n*(n-1)//2
    base = 0
    mod = base + 1
    for _ in range(0,max_iter):
        a[mod] = a[mod] - a[base]*np.dot(a[mod],a[base])
        a[mod] = a[mod]/np.linalg.norm(a[mod])

        if mod < n-1:
            mod += 1
        else:
            base += 1
            mod = base + 1
    return a



from unittest import TestCase
import random
import numpy as np
from math import cos, sin

from mgen.rotation_matrix_3d import _generate_matrix_XZX
from mgen.rotation_matrix_3d import _generate_matrix_XYX
from mgen.rotation_matrix_3d import _generate_matrix_YXY
from mgen.rotation_matrix_3d import _generate_matrix_YZY
from mgen.rotation_matrix_3d import _generate_matrix_ZYZ
from mgen.rotation_matrix_3d import _generate_matrix_ZXZ
from mgen.rotation_matrix_3d import _generate_matrix_XZY
from mgen.rotation_matrix_3d import _generate_matrix_XYZ
from mgen.rotation_matrix_3d import _generate_matrix_YXZ
from mgen.rotation_matrix_3d import _generate_matrix_YZX
from mgen.rotation_matrix_3d import _generate_matrix_ZYX
from mgen.rotation_matrix_3d import _generate_matrix_ZXY
from mgen import rotation_from_angles
from mgen import rotation_around_x
from mgen import rotation_around_y
from mgen import rotation_around_z
from mgen import rotation_around_axis
from mgen import rotation_from_angle_and_plane
from mgen import rotation_from_angle
from mgen import random_matrix

def is_close(m1, m2):
    np.testing.assert_allclose(m1, m2, atol=1.e-7)

def test_matrix_generation(func, single_rotations, name):
    rot_m = lambda a1, a2, a3: func(cos(a1), cos(a2), cos(a3), sin(a1), sin(a2), sin(a3))
    r1, r2, r3 = single_rotations
    rot_conv = lambda a1, a2, a3: rotation_from_angles((a1, a2, a3), name)

    for rand1, rand2, rand3 in np.random.uniform(-np.pi, np.pi, (10000, 3)):
        is_close(rot_m(rand1, rand2, rand3), r1(rand1).dot(r2(rand2)).dot(r3(rand3)))
        is_close(rot_m(rand1, rand2, rand3), rot_conv(rand1, rand2, rand3))

class TestRotations(TestCase):
    def test_rotation_2d(self):
        is_close(rotation_from_angle(0), np.eye(2))
        is_close(rotation_from_angle(np.pi/2), [[0, -1], [1, 0]])
        is_close(rotation_from_angle(np.pi), [[-1, 0], [0, -1]])
    def test_rotation_x(self):
        is_close(rotation_around_x(0), np.eye(3))
        is_close(rotation_around_x(np.pi/2), [[1, 0, 0], [0, 0, -1], [0, 1, 0]])
        is_close(rotation_around_x(np.pi), np.diag([1,-1,-1]))

    def test_rotation_y(self):
        is_close(rotation_around_y(0), np.eye(3))
        is_close(rotation_around_y(np.pi/2), [[0, 0, 1], [0, 1, 0], [-1, 0, 0]])
        is_close(rotation_around_y(np.pi), np.diag([-1, 1, -1]))

    def test_rotation_z(self):
        is_close(rotation_around_z(0), np.eye(3))
        is_close(rotation_around_z(np.pi/2), [[0, -1, 0], [1, 0, 0], [0, 0, 1]])
        is_close(rotation_around_z(np.pi), np.diag([-1, -1, 1]))


    def test_rotation_XZX(self):
        test_matrix_generation(_generate_matrix_XZX, (rotation_around_x, rotation_around_z, rotation_around_x), 'XZX')

    def test_rotation_XYX(self):
        test_matrix_generation(_generate_matrix_XYX, (rotation_around_x, rotation_around_y, rotation_around_x), 'XYX')

    def test_rotation_YXY(self):
        test_matrix_generation(_generate_matrix_YXY, (rotation_around_y, rotation_around_x, rotation_around_y), 'YXY')

    def test_rotation_YZY(self):
        test_matrix_generation(_generate_matrix_YZY, (rotation_around_y, rotation_around_z, rotation_around_y), 'YZY')

    def test_rotation_ZYZ(self):
        test_matrix_generation(_generate_matrix_ZYZ, (rotation_around_z, rotation_around_y, rotation_around_z), 'ZYZ')

    def test_rotation_ZXZ(self):
        test_matrix_generation(_generate_matrix_ZXZ, (rotation_around_z, rotation_around_x, rotation_around_z), 'ZXZ')

    def test_rotation_XZY(self):
        test_matrix_generation(_generate_matrix_XZY, (rotation_around_x, rotation_around_z, rotation_around_y), 'XZY')

    def test_rotation_XYZ(self):
        test_matrix_generation(_generate_matrix_XYZ, (rotation_around_x, rotation_around_y, rotation_around_z), 'XYZ')

    def test_rotation_YXZ(self):
        test_matrix_generation(_generate_matrix_YXZ, (rotation_around_y, rotation_around_x, rotation_around_z), 'YXZ')

    def test_rotation_YZX(self):
        test_matrix_generation(_generate_matrix_YZX, (rotation_around_y, rotation_around_z, rotation_around_x), 'YZX')

    def test_rotation_ZYX(self):
        test_matrix_generation(_generate_matrix_ZYX, (rotation_around_z, rotation_around_y, rotation_around_x), 'ZYX')

    def test_rotation_ZXY(self):
        test_matrix_generation(_generate_matrix_ZXY, (rotation_around_z, rotation_around_x, rotation_around_y), 'ZXY')

    def test_raises_if_sequence_unknown(self):
        with self.assertRaises(ValueError):
            rotation_from_angles((0, 0, 0), 'AXY')

    def test_rotation_around_axis(self):
        for random in (np.random.rand(10000) - 0.5) * 2 * np.pi:
            is_close(rotation_around_axis([1, 0, 0], random), rotation_around_x(random))
            is_close(rotation_around_axis([0, 1, 0], random), rotation_around_y(random))
            is_close(rotation_around_axis([0, 0, 1], random), rotation_around_z(random))
            # Test if non-normalised axis works, too
            is_close(rotation_around_axis([2, 0, 0], random), rotation_around_x(random))

        is_close(rotation_around_axis([1, 1, 1], 2*np.pi/3), rotation_from_angles([0, np.pi/2, np.pi/2], 'XYZ'))

    def test_accepts_tuple_list_and_array(self):
        angles_t = (0.1, 0.2, 0.3)
        angles_l = list(angles_t)
        angles_a = np.array(angles_l)

        np.testing.assert_allclose(
            rotation_from_angles(angles_t, 'YZY'),
            rotation_from_angles(angles_l, 'YZY'))

        np.testing.assert_allclose(
            rotation_from_angles(angles_t, 'YZY'),
            rotation_from_angles(angles_a, 'YZY'))

    def test_rotation_nd(self):
        for rand1 in np.random.uniform(-np.pi, np.pi, 100):
            # Test 2D case
            is_close(rotation_from_angle(rand1), rotation_from_angle_and_plane(rand1, (1,0), (0,1)))
            # Test 3D case
            is_close(rotation_around_z(rand1), rotation_from_angle_and_plane(rand1, (1,0,0), (0,1,0)))
            # Test normalisation of first parameter
            is_close(rotation_around_z(rand1), rotation_from_angle_and_plane(rand1, (2,0,0), (0,1,0)))
            # Test normalisation of second parameter
            is_close(rotation_around_z(rand1), rotation_from_angle_and_plane(rand1, (1,0,0), (0,2,0)))
            # Test normalisation of both parameters
            is_close(rotation_around_z(rand1), rotation_from_angle_and_plane(rand1, (2,0,0), (0,2,0)))

            is_close(rotation_around_x(rand1), rotation_from_angle_and_plane(rand1, (0,1,0), (0,0,1)))
            is_close(rotation_around_y(rand1), rotation_from_angle_and_plane(rand1, (0,0,1), (1,0,0)))

            # Test non orthogonal vectors
            is_close(rotation_around_x(rand1), rotation_from_angle_and_plane(rand1, (0, 1, 1), (0, 1, 2)))

            # Test generic properties of higher dimensional rotations

            # create two perpendicular random vectors
            dimension = random.randint(4,100)
            v1 = np.random.uniform(-1., 1., dimension)
            v1 = v1/np.linalg.norm(v1)
            v2 = np.random.uniform(-1., 1., dimension)
            v2 = v2/np.linalg.norm(v2)
            v2 = v2 - v2.dot(v1)*v1

            m = rotation_from_angle_and_plane(rand1, v1, v2)
            m_inv = rotation_from_angle_and_plane(-rand1, v1, v2)
            is_close(m.dot(m.T), np.eye(dimension))
            self.assertAlmostEqual(np.linalg.det(m), 1.0)
            is_close(m, m_inv.T)

        with self.assertRaises(ValueError):
            rotation_from_angle_and_plane(0., (1,0,0), (0,1,0,0))
        with self.assertRaises(ValueError):
            rotation_from_angle_and_plane(0., (1,0,0,0), (0,1,0))
        with self.assertRaises(ValueError):
            rotation_from_angle_and_plane(0., (1,0,0), (1,0,0))
        with self.assertRaises(ValueError):
            rotation_from_angle_and_plane(0., (0,0), (1,0))
        with self.assertRaises(ValueError):
            rotation_from_angle_and_plane(0., (1,0), (0,0))

    def test_random_matrix(self):
        for x in range(2,7):
            vecx = np.zeros(x, dtype=np.float64)
            vecx[0] = 1.0
            m = random_matrix(x)
            rotx = np.matmul(m,vecx)
            lengthx = np.sum(rotx*rotx)
            self.assertAlmostEqual(lengthx, 1.0)
        with self.assertRaises(ValueError):
            random_matrix(1)
